"""


"""


class Context(object):
    def __init__(self, base=None):
        super(Context, self).__init__()
        self._base = base
        self._values = dict()
        self._deleted = set()

    def get(self, name, default=None):
        try:
            return self.__getitem__(name)
        except KeyError:
            return default

    def update(self, **values):
        # ! can't use self._values.update(values) here:
        # we need to iterate so that self._deleted
        # is updated accordingly !
        for k, v in values.items():
            self[k] = v

    def __getitem__(self, name):
        try:
            return self._values[name]
        except KeyError:
            if self._base is None or name in self._deleted:
                raise KeyError(name)
            return self._base[name]

    def __setitem__(self, name, value):
        try:
            self._deleted.remove(name)
        except KeyError:
            pass
        self._values[name] = value

    def __delitem__(self, name):
        if name in self._deleted:
            raise KeyError(
                f'The name "{name}" has already been deleted in this context.'
            )
        try:
            del self._values[name]
        except KeyError:
            if self._base is not None:
                # Ensure their is something to delete
                try:
                    self._base[name]
                except KeyError:
                    raise
            else:
                raise
        self._deleted.add(name)

    def to_dict(self):
        d = {}
        if self._base is not None:
            d.update(self._base.to_dict())
        d.update(self._values)
        for key in self._deleted:
            del d[key]
        return d

    def edits(self):
        """
        Return a dict of added names, a dict of overridden names and a list of
        deleted names.
        """
        adds = dict()
        overs = dict()
        base = dict()
        if self._base is not None:
            base.update(self._base.to_dict())
        for k, v in self._values.items():
            try:
                base[k]
            except KeyError:
                adds[k] = v
            else:
                overs[k] = v
        return adds, overs, list(self._deleted)


class QAContext(Context):
    _KEY_ALLOW_AUTO_FIX = "SQARF::allow_auto_fix"
    _KEY_STOP_ON_FAIL = "SQARF::stop_on_fail"

    def stop_on_fail(self):
        return self.get(self._KEY_STOP_ON_FAIL, False)

    def set_stop_on_fail(self, stop_on_fail):
        """
        `stop_on_fail` can be True, False, or None for default behavior
        """
        if stop_on_fail is None:
            del self[self._KEY_STOP_ON_FAIL]
        else:
            self[self._KEY_STOP_ON_FAIL] = stop_on_fail

    def allow_auto_fix(self):
        return self.get(self._KEY_ALLOW_AUTO_FIX, False)

    def set_allow_auto_fix(self, allow_auto_fix):
        """
        `allow_auto_fix` can be True, False, or None for default behavior
        """
        if allow_auto_fix is None:
            del self[self._KEY_ALLOW_AUTO_FIX]
        else:
            self[self._KEY_ALLOW_AUTO_FIX] = allow_auto_fix



"""




"""
import os
import json

from .context import Context
from .qatest import QATest
from . import html_tree_export


class _SessionRootTest(QATest):

    SESSION = None

    @classmethod
    def get_sub_test_types(cls):
        return cls.SESSION._test_types

    def can_fix(self, context):
        return False, "Can't fix the whole world :/"


class Session(object):
    def __init__(self):
        super(Session, self).__init__()

        # This is sooooooooooo hacky !!! ^o^
        # I'll burn in hell for this :p
        class Root(_SessionRootTest):
            SESSION = self

        self._test_types = []
        self._root_test_type = Root

        self._stop_on_fail = None
        self._allow_auto_fix = None
        self._context = Context(None)

        self._runs = []

    def register_test_types(self, test_types):
        self._test_types.extend(test_types)

    def register_tests_from_file(self, filename):
        """
        WARNING: only use this on trusted files !

        The python code in `filename` file must declare a `get_root_tests()`
        function returning a list of test types to register.

        The code can also use the gloable name `SESSION` which contains
        the session being configured.
        """
        if not os.path.isfile(filename):
            raise ValueError(f"{filename} is not a file !")

        with open(filename, "rb") as fp:
            content = fp.read()

        self.register_tests_from_string(content, filename)

    def register_tests_from_string(self, code_string, virtual_filename):
        """
        WARNING: only use this with trusted data ! The `code_string` has
        the ability to wipe your server clean or tell anything to your
        mother in law !

        The python code in `code_string` string must declare a `get_root_tests()`
        function returning a list of test types to register.

        The code can also use the gloable name `SESSION` which contains
        the session being configured.

        The `virtual_filename` string will be used as the "source filename" in
        reports and some exceptions.
        """
        compiled = compile(code_string, virtual_filename, "exec")
        namespace = {"SESSION": self, "__filename__": virtual_filename}
        exec(compiled, namespace, namespace)
        getter_name = "get_root_tests"
        getter = namespace.get(getter_name)
        if getter is None:
            raise Exception(
                f'Could not find a "{getter_name}()" function in {virtual_filename}. '
                "No test registered."
            )

        try:
            test_types = getter()
        except Exception as err:
            import traceback

            traceback.print_exc()
            raise Exception(
                f'Error executing "{getter_name}()" from {virtual_filename}: {err}'
            )

        for TestType in test_types:
            TestType._forced_filename = virtual_filename
        self._test_types.extend(test_types)

    def context_set(self, **values):
        self._context.update(**values)

    def run(self):
        root_test = self._root_test_type()
        self._runs.append(root_test)

        # use a copy of the context, tests will modify it:
        context = Context(self._context)

        # configure the context as requested:
        if self._stop_on_fail is not None:
            root_test.set_stop_on_fail(context, self._stop_on_fail)
        if self._allow_auto_fix is not None:
            root_test.set_allow_auto_fix(context, self._allow_auto_fix)

        # run all the tests:
        result = root_test.run(context)
        return result

    def to_lines(self):
        lines = []
        for run in self._runs:
            lines.extend(run.to_lines())
        return lines

    def to_dict_list(self):
        ret = []
        for run in self._runs:
            ret.append(run.to_dict())
        return ret

    def to_json(self):
        as_dict_list = self.to_dict_list()
        return json.dumps(as_dict_list)

    def to_json_file(self, filename, force_overwrite=False):
        as_list = self.to_dict_list()
        if os.path.exists(filename) and not force_overwrite:
            raise ValueError(f"filename {filename} already exists !")
        with open(filename, "w") as fp:
            json.dump(as_list, fp)

    def to_html_tree(self, filename, force_overwrite=False):
        session_json = self.to_json()
        content = html_tree_export.html_tree(session_json)
        if os.path.exists(filename) and not force_overwrite:
            raise ValueError(f"filename {filename} already exists !")
        with open(filename, "w") as fp:
            fp.write(content)



"""



"""
import time
import traceback
import inspect

from .context import QAContext


class QAResult(object):
    class STATUS:
        ERROR = "ERROR"
        NOT_RAN = "NOT RAN"
        SKIPPED = "SKIPPED"
        IRRELEVANT = "IRRELEVANT"
        PASSED = "PASSED"
        FAILED = "FAILED"
        FIXED = "FIXED"
        FIX_FAILED = "FIX FAILED"

    @classmethod
    def ERROR(cls, why, err, trace):
        return cls(cls.STATUS.ERROR, why, err, trace)

    @classmethod
    def NOT_RAN(cls):
        return cls(cls.STATUS.NOT_RAN, "Not yet ran.")

    @classmethod
    def SKIPPED(cls, why):
        return cls(cls.STATUS.SKIPPED, why)

    @classmethod
    def IRRELEVANT(cls, why):
        return cls(cls.STATUS.IRRELEVANT, why)

    @classmethod
    def PASSED(cls, why):
        return cls(cls.STATUS.PASSED, why)

    @classmethod
    def FAILED(cls, why):
        return cls(cls.STATUS.FAILED, why)

    @classmethod
    def FIXED(cls, why):
        return cls(cls.STATUS.FIXED, why)

    @classmethod
    def FIX_FAILED(cls, why, previous_result):
        result = cls(cls.STATUS.FIX_FAILED, why)
        result._before_fix_result = previous_result
        return result

    def __init__(self, status, summary, err=None, trace=None):
        super(QAResult, self).__init__()
        self._status = status
        self._summary = summary
        self._error = err
        self._trace = trace
        self._before_fix_result = None
        self._timestamp = time.time()

    def failed(self):
        return self._status in (
            self.STATUS.FAILED,
            self.STATUS.FIX_FAILED,
            self.STATUS.ERROR,
        )

    def passed(self):
        return self._status in (
            self.STATUS.SKIPPED,
            self.STATUS.PASSED,
            self.STATUS.FIXED,
            self.STATUS.IRRELEVANT,
        )

    def not_applicable(self):
        return self._status in (
            self.STATUS.SKIPPED,
            self.STATUS.NOT_RAN,
            self.STATUS.IRRELEVANT,
        )

    def to_dict(self):
        before_fix = None
        if self._before_fix_result is not None:
            before_fix = self._before_fix_result.to_dict()
        return dict(
            timestamp=self._timestamp,
            status=self._status,
            summary=str(self._summary),
            error=self._error and str(self._error) or None,
            trace=self._trace,
            before_fix=before_fix,
        )

    def __bool__(self):
        return self.passed()

    def __str__(self):
        s = "{} ({})".format(self._status, self._summary)
        if self._error is not None:
            s += ': "{!s}"'.format(self._error)
            s += "\n" + self._trace
        return s

    def __add__(self, other):
        if not isinstance(other, self.__class__):
            raise TypeError()

        summary = self._summary
        status = self._status

        if not other.not_applicable():
            if self.failed():
                status = self._status
            elif other.failed():
                status = other._status
                summary = "Sub Test Failed"
        return self.__class__(status, summary)


class QATest(object):

    _forced_filename = None

    @classmethod
    def test_filename(cls):
        if cls._forced_filename is not None:
            return cls._forced_filename
        else:
            return inspect.getfile(cls)

    @classmethod
    def test_name(cls):
        """
        Returns the name for this QATest.
        May be overridden to return any string.

        Default is to use the class name.
        """
        return cls.__name__

    @classmethod
    def test_description(cls):
        """
        Returns short and long description for this QATest.
        May be overridden to return two strings.

        The short description should not contain "\\n".
        The long description may be an empty string.

        Default is to use the class docstring:
        first line is short description and the remaining is
        the long description.
        """
        doc = (cls.__doc__ or "").strip()
        if not doc:
            return None, None
        try:
            short, long = doc.split("\n", 1)
        except ValueError:
            return doc, None
        else:
            nice_long = ""
            for line in long.split("\n\n"):
                for wrapped in line.split("\n"):
                    nice_long += wrapped.strip() + " "
                nice_long += "\n"

        return short.strip(), nice_long.strip()

    @classmethod
    def get_sub_test_types(cls):
        return []

    def __init__(self):
        super(QATest, self).__init__()
        self._timestamp = time.time()
        self._context = None
        self._result = None
        self._sub_tests = []  # QATest instances, filled at run()

    def relevant_for(self, context):
        """
        Returns bool, str: relevant for this context, detail message
        `context` may be modified.
        """
        return True, "Default Relevance"

    def can_fix(self, context):
        return False, "No Fix for this."

    def fix(self, context):
        """
        Must return bool, str:
            boo: True if fix succeeded, False if not.
            str: summary (describes why it succeeded or not)
        """
        raise NotImplementedError(self)

    def _do_test(self, context):
        """
        Must return bool, str:
            boo: True if test passed, False if test failed.
            str: summary (tells why it failed or passed)

        Default is to return True with empty summary

        Do not attempt to use `self.fix()` here, it will be called
        as needed later on.
        """
        return True, ""

    def _store_sub_test(self, test):
        self._sub_tests.append(test)

    def _results_sum(self, result):
        """
        Return the sum of `result` and all sub-tests results.
        """
        sub_results = [t._result for t in self._sub_tests]
        return sum(sub_results, result)

    def _run_sub_tests(self, context):
        """
        Build sub-test instances, run them and store
        results.
        """
        stop_subs = False
        for TestType in self.get_sub_test_types():
            # Build Test
            try:
                test = TestType()
            except Exception as err:
                trace = traceback.format_exc()
                result = QAResult.ERROR(
                    "Error creating sub-test {}. Aborting.".format(TestType), err, trace
                )
                self._result = result
                return

            self._store_sub_test(test)

            if stop_subs:
                test._result = QAResult.SKIPPED("Test stopped after fail.")
                continue

            # Check Relevance
            try:
                relevant, why = test.relevant_for(context)
            except Exception as err:
                trace = traceback.format_exc()
                result = QAResult.ERROR("Error checking relevance.", err, trace)
                test._result = result
            else:
                if not relevant:
                    result = QAResult.IRRELEVANT(why)
                    test._result = result
                    continue

            # Run Test and its Sub-Tests
            result = test.run(context)
            if result.failed() and context.stop_on_fail():
                stop_subs = True
                continue

    def run(self, base_context):
        """
        Runs this test and all its sub-tests.
        Return bool, str: passed, summary
        """
        if self._context is not None:
            raise Exception("Test already performed.")

        # backup the context for later diff
        self._context = QAContext(base_context)

        # run this test
        try:
            passed, why = self._do_test(self._context)
        except Exception as err:
            trace = traceback.format_exc()
            result = QAResult.ERROR("Could not run test.", err, trace)
            self._result = result
            return result
        else:
            if passed:
                result = QAResult.PASSED(why)
            else:
                result = QAResult.FAILED(why)
        self._result = result

        # fix if needed, possible and allowed:
        if (
            result.failed()
            and self._context.allow_auto_fix()
            and self.can_fix(self._context)
        ):
            fixed, why = self.fix(self._context)
            if fixed:
                result = QAResult.FIXED(why)
            else:
                result = QAResult.FIX_FAILED(why, self._result)
            self._result = result

        # cancel sub-tests if needed:
        if result.failed() and self._context.stop_on_fail():
            return result

        # run all sub-tests
        self._run_sub_tests(self._context)

        # Update our result with sub-result:
        self._result = self._results_sum(result)

        # Return our result for convenience
        return self._result

    def to_lines(self, indent=0):
        indent_str = indent * "  "
        lines = ["{}{}:{}".format(indent_str, self.test_name(), self._result)]
        for test in self._sub_tests:
            lines.extend(test.to_lines(indent + 1))
        return lines

    def to_dict(self):
        """Returns a dict containing info for this test and all sub tests"""
        short_description, long_description = self.test_description()

        context_dict = {}
        context_edits = None
        if self._context is not None:
            context_dict = self._context.to_dict()
            context_edits = self._context.edits()

        qualname = "{}.{}".format(
            self.__class__.__module__,
            self.__class__.__name__,
        )
        return dict(
            test_type=qualname,
            test_filename=self.test_filename(),
            test_name=self.test_name(),
            short_description=short_description,
            long_description=long_description,
            timestamp=self._timestamp,
            sub_tests=[t.to_dict() for t in self._sub_tests],
            context_edits=context_edits,
            context=context_dict,
            result=self._result.to_dict(),
        )



import os

from setuptools import setup, find_packages
import versioneer

readme = os.path.normpath(os.path.join(__file__, "..", "README.md"))
with open(readme, "r") as fh:
    long_description = fh.read()


setup(
    cmdclass=versioneer.get_cmdclass(),
    name="sqarf",
    version=versioneer.get_version(),
    description=("Simple Quality Assurance Reports and Fixes"),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/kabaretstudio/sqarf",
    author='Damien "dee" Coureau',
    author_email="dee909@gmail.com",
    license="LGPLv3+",
    classifiers=[
        # 'Development Status :: 1 - Planning',
        # 'Development Status :: 2 - Pre-Alpha',
        # "Development Status :: 3 - Alpha",
        "Development Status :: 4 - Beta",
        # 'Development Status :: 5 - Production/Stable',
        # 'Development Status :: 6 - Mature',
        # 'Development Status :: 7 - Inactive',
        # 'Topic :: System :: Shells',
        "Intended Audience :: Developers",
        # 'Intended Audience :: End Users/Desktop',
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: "
        "GNU Lesser General Public License v3 or later (LGPLv3+)",
    ],
    keywords="kabaret blender b3d pipeline dataflow workflow",
    install_requires=[],
    extras_require={
        'dev': ['pytest', 'pytest-cov', 'flake8', 'black'],
    },
    # python_requires=">=3.7",
    packages=find_packages("src"),
    package_dir={"": "src"},
)



# flake8: noqa
import time
import json


HEAD_ICONS = """
<!-- icons from https://remixicon.com/ -->
<link href="https://cdn.jsdelivr.net/npm/remixicon@2.5.0/fonts/remixicon.css" rel="stylesheet">
"""

HEAD_CSS_TOOLBAR = """
  <style type="text/css">
  /* Tools CSS */

    /* Add a black background color to the top navigation */
    .top_tools, .top_tools a{
        background-color: #333;
        overflow: hidden;
        display: flex;
        align-items: center;
    }

    /* toggleable resize when hidden-toggle button resizes */
    .toggleable {
        padding: 5px;
    }

    /* Style the links inside the top bar */
    .top_tools a {
        float: left;
        color: #f2f2f2;
        text-align: center;
        padding: 14px 16px;
        text-decoration: none;
        font-size: 17px;
    }

    /* Add a color to the active tools */
    .top_tools a.active {
        background-color: #4CAF50;
        color: white;
    }

    /* Change the color of links on hover */
    .top_tools a:hover {
        background-color: #ddd;
        color: black;
    }
  </style>
"""

HEAD_CSS_TOOLTIPS = """
  <style type="text/css">
  /* Tooltips CSS */
    .tooltip {
    position: relative;
    display: inline-block;
    }

    .tooltip .tooltiptext {
    visibility: hidden;
    //width: 120px;
    background-color: black;
    color: #fff;
    //text-align: center;
    border-radius: 6px;
    padding: 10px 10px 5px 5px;
    position: absolute;
    z-index: 1;
    top: -5px;
    left: 110%;
    }

    .tooltip .tooltiptext::after {
    content: "";
    position: absolute;
    top: 20px;
    right: 100%;
    margin-top: -5px;
    border-width: 5px;
    border-style: solid;
    border-color: transparent black transparent transparent;
    }
    .tooltip:hover .tooltiptext {
    visibility: visible;
    }
  </style>
"""

HEAD_CSS_TABLETREE = """
  <style type="text/css">
  /* TableTree CSS */

  table.tree {
  	border-collapse: collapse;
    width: 100%;
  }

  table.tree tr {
  }

  table.tree tr:hover {
      text-shadow: 2px 2px 2px #00000036;
  }

  table.tree td {
  }

  table.tree td:first-child {
  }
  </style>
"""

HEAD_CSS_LOOK = """
  <style type="text/css">
  /* Look CSS */

    body {
      font-family: Arial;
    }

    .hfill {
        width: 100%;
        background-color=red;
    }
    
    .status {
      padding: 5;
      margin: 0 -5px 0 0;
    }

    .ERROR {
        background-color: red;
        color: cyan;
    }
    .FAILED {
      background-color: fdd;
      color: maroon;
    }
    .PASSED {
      background-color: dfd;
      color: green;
    }
    .IRRELEVANT {
      background-color: #bbb;
      color: #eee;
    }
    .NA {
      background-color: grey;
      color: white;
    }
  </style>
"""

HEAD = (
    "<head>"
    "<title>SQARF Report</title>" + HEAD_ICONS + HEAD_CSS_TOOLBAR + HEAD_CSS_TOOLTIPS
    # + HEAD_CSS_TREE
    + HEAD_CSS_TABLETREE + HEAD_CSS_LOOK + "</head>"
)

BODY_TOOLBAR = """
<div class="top_tools">
    <!-- <a class="button remove-class" affected=".with_children" to_remove="open"><i class="ri-add-box-line"></i>All</a> -->
    <!-- <a class="button add-class" affected=".with_children" to_add="open"><i class="ri-checkbox-indeterminate-line"></i>All</a> -->

    <!-- <a class="button hidden-toggle active" to_toggle=".test-name"><i class="ri-checkbox-fill ri-1x"></i><span>Show Test Name</span></a> -->
    <!-- <a class="button hidden-toggle active" to_toggle=".test-summary"><i class="ri-checkbox-fill ri-1x"></i><span>Summaries</span></a> -->
    <a class="button hidden-toggle active" to_toggle=".info-description"><i class="ri-checkbox-fill ri-1x"></i> Descriptions</a>
    <a class="button hidden-toggle active" to_toggle=".info-time"><i class="ri-checkbox-fill ri-1x"></i>Show Run Times</a>
    <a class="button hidden-toggle active" to_toggle=".info-source"><i class="ri-checkbox-fill ri-1x"></i>Show Test Source</a>
<!--    <a class="button hidden-toggle active" to_toggle=".PASSED"><i class="ri-checkbox-fill ri-1x"></i>Hide Passed</a> -->
<!--   <a class="button hidden-toggle active" to_toggle=".IRRELEVANT"><i class="ri-checkbox-fill ri-1x"></i>Show Irrelevants</a> -->
    <!-- <a class="button hidden-toggle active" to_toggle=".info-traceback"><i class="ri-checkbox-fill ri-1x"></i>Traces</a> -->
</div>
"""

TOP = "<html>" + HEAD + "<body>" + BODY_TOOLBAR

JS_TOOLBAR = """ 
    /* button click */
    var buttons = document.querySelectorAll('.button, .button *');
    for (var i = 0; i < buttons.length; i++){
        buttons[i].addEventListener('click', function(e) {
            e.stopPropagation(); /* needed to avoid double trigger on image click */
            var b = e.srcElement.closest('a');

            // Visibility toggle button
            if(b.classList.contains('hidden-toggle')) {
                b.classList.toggle('active');
                b.querySelector('i').classList.toggle('ri-checkbox-blank-line');
                b.querySelector('i').classList.toggle('ri-checkbox-fill');
                var toggleable = document.querySelectorAll('.toggleable'+b.getAttribute('to_toggle'));
                for(var i=0; i < toggleable.length; i++){
                    toggleable[i].hidden = ! toggleable[i].hidden;
                }
            }

            // Class add button
            if(b.classList.contains('add-class')) {
                var affected = document.querySelectorAll(b.getAttribute('affected'));
                for(var i=0; i < affected.length; i++){
                    affected[i].classList.remove(b.getAttribute('to_add'));
                }
            }

            // Class remove button
            if(b.classList.contains('remove-class')) {
                var affected = document.querySelectorAll(b.getAttribute('affected'));
                for(var i=0; i < affected.length; i++){
                    affected[i].classList.add(b.getAttribute('to_remove'));
                }
            }
        });
    }

    /* click toggle buttons to set default state */
    var hbs = document.querySelectorAll('.hidden-toggle')
    for(var i = 0; i < hbs.length; i++){
        hbs[i].dispatchEvent(new Event('click'));
    }
"""

JS_TABLETREE = """

    /* Table Tree */
	var collect_sibblings = function (elem, skip_func, break_func) {
		var collected = [];
		elem = elem.nextElementSibling;
		while (elem) {
			console.log('?', elem, elem.getAttribute('data-depth'), skip_func(elem), break_func(elem));
			if (break_func(elem)) break;
			if (!skip_func(elem)){
                console.log('+', elem)
                collected.push(elem);
            };
			elem = elem.nextElementSibling;
		}
		return collected;
	};
    var findAllSubs = function (tr) {
        var depth = parseInt(tr.getAttribute('data-depth'));
        var break_if = function (el) {
            return el.getAttribute('data-depth') <= depth;
        }
        var skip_if = function (el) {
            return false;
        }
        return collect_sibblings(tr, skip_if, break_if);
    };
    var findDirectSubs = function (tr) {
        var depth = parseInt(tr.getAttribute('data-depth'));
        var break_if = function (el) {
            return el.getAttribute('data-depth') <= depth;
        }
        var skip_if = function (el) {
            return el.getAttribute('data-depth') > depth+1;
        }
        return collect_sibblings(tr, skip_if, break_if);
    };
    var collapse = function(tr) {
        var subs = findAllSubs(tr);
        tr.classList.add('collapsed');
        for(i=0; i<subs.length; i++){
            subs[i].hidden = true;
            if(!subs[i].classList.contains('collapsed')){
                collapse(subs[i]);
            }
        }
    };
    var expand = function(tr) {
        var subs = findDirectSubs(tr);
        tr.classList.remove('collapsed');
        console.log('expand', tr, subs);
        for(i=0; i<subs.length; i++){
            subs[i].hidden = false;
        }
    }

    // init first tds indent
    var first_tds = document.querySelectorAll('table.tree td.collapser');
    for(var i = 0; i < first_tds.length; i++){
        var indent = first_tds[i].parentElement.getAttribute('data-depth')*50;
        first_tds[i].setAttribute("style", "border-left: "+indent+"px solid #fff");
    };
    
    // set .toggle click
    var collapsers = document.querySelectorAll('table.tree .collapser');
    for(var i = 0; i < collapsers.length; i++){
        collapsers[i].addEventListener('click', function(e) {
            var tr = e.srcElement.closest('tr'); //Get <tr> parent of tr collapser 
            if(tr.classList.contains('collapsed')) {
                expand(tr);
            }else{
                collapse(tr);
            }
        });
    };

"""

BOTTOM = (
    '<script type="text/javascript">'
    + JS_TOOLBAR
    + JS_TABLETREE
    + "</script>"
    + "</body></html>"
)


def test_to_html_tree(test, root_timestamp, indent=0):
    """
    Returns the html table-tree for the given test json export.
    """
    S = indent * "  "
    lines = []
    test_name = test["test_name"]
    short = test["short_description"]
    long = test["long_description"]
    start_time = test["timestamp"] - root_timestamp
    run_time = test["result"]["timestamp"] - test["timestamp"]
    result = test["result"]
    status = result["status"]
    summary = result["summary"] or ""
    error = result["error"]
    trace = result["trace"]

    lines = []
    lines.append(f'{S}<tr data-depth="{indent}" class="">')

    # Test name + tooltip status and summary
    lines.append(f'{S}  <td class="collapser status {status}">')
    if test["sub_tests"]:
        lines.append(f'{S}<i class="ri-flask-fill"></i>')
    else:
        lines.append(f'{S}<i class="ri-test-tube-fill"></i>')
    lines.append(f'{S}    <span class="tooltip">{test_name}')
    lines.append(f'{S}      <span class="tooltiptext">')
    lines.append(f'{S}        <span class="status {status}">{status}</span>')
    lines.append(f"{S}        <pre>{summary}</pre>")
    if error:
        lines.append(f"{S}        <pre>{error}</pre>")
        lines.append(f"{S}        <pre>{trace}</pre>")

    lines.append(f"{S}        </span>")
    lines.append(f"{S}      </span>")
    lines.append(f"{S}    </td>")

    # Test short description + tooltip long description
    lines.append(f'{S}    <td class="{status} toggleable info-description">')
    lines.append(f'{S}      <span class="tooltip">{short or ""}')
    if long:
        lines.append(f'{S}      <span class="tooltiptext">')
        lines.append(f"{S}        <pre>{long}</pre>")
        lines.append(f"{S}    </span>")
    lines.append(f"{S}    </td>")

    # Run times
    lines.append(
        f'{S}    <td class="{status} toggleable info-time">start @+{start_time:.02f}, {run_time:.02f}s long.</td>'
    )

    # Test type + tooltip source file
    lines.append(f'{S}    <td class="{status} toggleable info-source">')
    lines.append(f'{S}      <span class="tooltip">{test["test_type"]}')
    lines.append(f'{S}      <span class="tooltiptext">')
    lines.append(f'{S}        <pre>{test["test_filename"]}</pre>')
    lines.append(f"{S}    </span>")
    lines.append(f"{S}    </td>")

    lines.append(f"{S}</tr>")

    # sub tests:
    for sub_test in test["sub_tests"]:
        lines.extend(test_to_html_tree(sub_test, root_timestamp, indent + 1))

    return lines


def html_tree(session_json_dumps):
    run_dicts = json.loads(session_json_dumps)
    lines = []
    for run in run_dicts:
        title = "Results on {}".format(time.ctime(run["timestamp"]))
        lines.append(f'<h3 class="session-run">{title}</h3>')
        lines.append('<table class="tree">')
        lines.extend(test_to_html_tree(run, run["timestamp"]))
        lines.append("</table>")
    lines = "\n".join(lines)
    return TOP + lines + BOTTOM
